/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/is_domain_ok_for_update.hh"

#include <libfred/object/object_states_info.hh>
#include <libfred/object_state/lock_object_state_request_lock.hh>
#include <libfred/object_state/perform_object_state_request.hh>
#include <libfred/registrable_object/domain/domain_uuid.hh>

#include <liblog/liblog.hh>

namespace Fred {
namespace Akm {
namespace Database {
namespace Impl {

namespace {

bool domain_state_policy_ok(const LibFred::OperationContext& _ctx, const LibFred::InfoDomainData& _info_domain_data)
{
    LibFred::LockObjectStateRequestLock(_info_domain_data.id).exec(_ctx);
    LibFred::PerformObjectStateRequest(_info_domain_data.id).exec(_ctx);
    const LibFred::ObjectStatesInfo domain_states(LibFred::GetObjectStates(_info_domain_data.id).exec(_ctx));

    if (domain_states.presents(LibFred::Object_State::server_blocked) ||
        domain_states.presents(LibFred::Object_State::delete_candidate))
    {
        LIBLOG_INFO("domain state policy error occurred, can't update.");
        return false;
    }
    return true;
}

} // namespace Fred::Akm::Database::Impl::{anonymous}

} // namespace Fred::Akm::Database::Impl
} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred

bool Fred::Akm::Database::Impl::is_domain_ok_for_update(
        const LibFred::OperationContext& _ctx,
        const LibFred::InfoDomainData& _info_domain_data,
        const DomainUuid& _domain_uuid,
        const KeysetMetadata& _keyset_metadata)
{
    const auto uuid_from_db = DomainUuid(::Util::get_raw_value_from(_info_domain_data.uuid));
    if (uuid_from_db != _domain_uuid)
    {
        LIBLOG_DEBUG("domain uuid from db {} does not match one from cdnskey-processor {}",
                to_string(uuid_from_db),
                to_string(_domain_uuid));
        return false;
    }
            
    const bool domain_zone_is_automatically_managed =
            _keyset_metadata.zones.find(_info_domain_data.zone.handle) !=
            _keyset_metadata.zones.end();
    if (!domain_zone_is_automatically_managed)
    {
        LIBLOG_DEBUG("domain zone {} is not managed", _info_domain_data.zone.handle);
        return false;
    }

    if (_info_domain_data.nsset.isnull())
    {
        LIBLOG_DEBUG("domain has empty nsset");
        return false;
    }

    if (!domain_state_policy_ok(_ctx, _info_domain_data))
    {
        LIBLOG_INFO("domain state policy error occurred");
        return false;
    }
    return true;
}
