/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg.hh"
#include "src/database/check_configuration.hh"
#include "src/exception.hh"
#include "src/export_scan_tasks.hh"
#include "src/update.hh"

#include "libfred/db_settings.hh"
#include "liblog/liblog.hh"
#include "liblog/log.hh"

#include <cstdlib>
#include <iostream>

namespace {

auto severity_to_level(Fred::Akm::Cfg::LogSeverity severity)
{
    using namespace Fred::Akm;
    switch (severity)
    {
        case Cfg::LogSeverity::critical:
            return LibLog::Level::critical;
        case Cfg::LogSeverity::error:
            return LibLog::Level::error;
        case Cfg::LogSeverity::warning:
            return LibLog::Level::warning;
        case Cfg::LogSeverity::info:
            return LibLog::Level::info;
        case Cfg::LogSeverity::debug:
            return LibLog::Level::debug;
        case Cfg::LogSeverity::trace:
            return LibLog::Level::trace;
    }
    return LibLog::Level::trace;
}

std::string make_connect_string(const Fred::Akm::Cfg::Options::Database& option)
{
    using namespace Fred::Akm;
    std::string result;
    if (option.host != boost::none)
    {
        result = "host=" + *option.host;
    }
    if (option.port != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "port=" + std::to_string(*option.port);
    }
    if (option.user != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "user=" + *option.user;
    }
    if (option.dbname != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "dbname=" + *option.dbname;
    }
    if (option.password != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "password=" + *option.password;
    }
    return result;
}

auto setup_logging(const Fred::Akm::Cfg::Options::Log& log_options)
{
    LibLog::LogConfig log_config;
    struct Result
    {
        bool log_enabled = false;
    } result;
    if (log_options.console != boost::none)
    {
        LibLog::Sink::ConsoleSinkConfig config;
        const auto& option = *log_options.console;
        config.set_level(severity_to_level(option.min_severity))
            .set_output_stream(LibLog::Sink::ConsoleSinkConfig::OutputStream::stderr)
            .set_color_mode(LibLog::ColorMode::automatic);
        log_config.add_sink_config(config);
        result.log_enabled = true;
    }
    if (log_options.file != boost::none)
    {
        const auto& option = *log_options.file;
        LibLog::Sink::FileSinkConfig config{option.file_name};
        config.set_level(severity_to_level(option.min_severity));
        log_config.add_sink_config(config);
        result.log_enabled = true;
    }
    if (log_options.syslog != boost::none)
    {
        const auto& option = *log_options.syslog;
        LibLog::Sink::SyslogSinkConfig config;
        if (!option.ident.empty())
        {
            config.set_syslog_ident(option.ident);
        }
        config.set_level(severity_to_level(option.min_severity));
        config.set_syslog_facility(option.facility);
        log_config.add_sink_config(config);
        result.log_enabled = true;
    }
    if (result.log_enabled)
    {
        LibLog::Log::start<LibLog::ThreadMode::multi_threaded>(log_config);
    }
    return result;
}

void setup_database(const Fred::Akm::Cfg::Options::Database& cfg)
{
    ::Database::emplace_default_manager<::Database::StandaloneManager>(make_connect_string(cfg));
}

Fred::Akm::KeysetMetadata make_keyset_metadata(const Fred::Akm::Cfg::Options::Akm& cfg)
{
    return {cfg.handle_suffix_size,
            cfg.handle_prefix,
            cfg.registrar,
            cfg.tech_contact,
            {begin(cfg.zones), end(cfg.zones)}};
}

int import_command(const Fred::Akm::Cfg::Options::Import& cmd)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, cmd.scan_type == Fred::Akm::ExportScanType::all ? "import"
                                                                                        : "import-" + to_string(cmd.scan_type));
    LIBLOG_INFO("export domains to scan");
    const auto& cfg = Fred::Akm::Cfg::Options::get();
    export_scan_tasks(
            cmd.scan_type,
            make_keyset_metadata(cfg.akm),
            Fred::Akm::SupervisorConnectionString{cfg.cdnskey_processor.supervisor_connection_string});
    return EXIT_SUCCESS;
}

int update_command(const Fred::Akm::Cfg::Options::Update& cmd)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, cmd.dry_run ? "update-dry-run"
                                                    : "update");
    LIBLOG_INFO("import acceptable cdnskey records");
    const auto& cfg = Fred::Akm::Cfg::Options::get();
    update_acceptable_keysets(
            make_keyset_metadata(cfg.akm),
            Fred::Akm::SupervisorConnectionString{cfg.cdnskey_processor.supervisor_connection_string},
            Fred::Akm::Logger::ConnectionString{cfg.logger.logger_connection_string},
            cmd.dry_run ? Fred::Akm::DryRun::yes
                        : Fred::Akm::DryRun::no);
    return EXIT_SUCCESS;
}

}//namespace {anonymous}

int main(int argc, char* argv[])
{
    using namespace Fred::Akm;
    try
    {
        Cfg::Options::init(argc, argv);
    }
    catch (const Cfg::AllDone& e)
    {
        std::cout << e.what() << std::endl;
        return EXIT_SUCCESS;
    }
    catch (const Cfg::Exception& e)
    {
        std::cerr << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const ConfigurationError& e)
    {
        std::cerr << "Configuration problem: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }

    if (setup_logging(Cfg::Options::get().log).log_enabled)
    {
        LIBLOG_DEBUG("fred-akm-ng configuration successfully loaded");
    }
    else
    {
        std::cout << "fred-akm-ng configuration successfully loaded" << std::endl;
    }
    try
    {
        setup_database(Cfg::Options::get().database);
        Fred::Akm::Database::check_configuration(make_keyset_metadata(Cfg::Options::get().akm));
        return run(Cfg::Options::get().command, import_command, update_command);
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        LIBLOG_ERROR("Exception caught: {}", e.what());
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        LIBLOG_ERROR("Unexpected exception caught");
    }
    return EXIT_FAILURE;
}
